/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Nickolay Ponomarev.
 *
 * Contributor(s):
 *  Nickolay Ponomarev <asqueella@gmail.com> (Original author)
 *  Malte Kraus <firefox@maltekraus.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;

var globalObject = this;


function log(msg) {
  /*msg = "Adblock Filters:\t" + msg;
  var consoleService = Cc["@mozilla.org/consoleservice;1"]
                         .getService(Ci.nsIConsoleService);
  consoleService.logStringMessage(msg);
  //dump(msg + "\n");
  return msg;*/
}
function makeURI(uri) {
  var ioservice = Cc["@mozilla.org/network/io-service;1"]
                   .getService(Ci.nsIIOService);
  return ioservice.newURI(uri, null, null);
}

Cc["@mozilla.org/moz/jssubscript-loader;1"]
  .getService(Ci.mozIJSSubScriptLoader)
  .loadSubScript("chrome://global/content/nsUserSettings.js");

function AdblockFilterService() {
  this.wrappedJSObject = this;
  this._registerPrefObserver();
}
AdblockFilterService.prototype = {
  classID: Components.ID("{76c938ef-4597-2f95-4f1a-fad72f7ecb6c}"),
  contractID: "@adblockfilters.mozdev.org/service;1",
  classDescription: "Adblock Plus Filter Uploader service",

  _closeObserverRegistered: false,
  _registerCloseObserver: function() {
    // does not work in Songbird which uses nsIAppStartup to kill the app
    if(this._closeObserverRegistered)
      return;
    this._closeObserverRegistered = true;

    Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService)
      .addObserver(this, "domwindowclosed", false);
  },
  _unregisterCloseObserver: function() {
    if(!this._closeObserverRegistered)
      return;
    this._closeObserverRegistered = false;

    Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService)
      .removeObserver(this, "domwindowclosed");
  },
  _registerPrefObserver: function() {
    var prefService = Cc["@mozilla.org/preferences-service;1"]
                        .getService(Ci.nsIPrefService);
    this._branch = prefService
                           .getBranch("extensions.adblockfilters.uploadOnExit");
    if(this._branch.getBoolPref("")) {
      this._registerCloseObserver();
    }
    this._branch.QueryInterface(Ci.nsIPrefBranch2);
    this._branch.addObserver("", this, false);
  },

  _unregisterPrefObserver: function() {
    if(!this._branch)
      return;
    this._branch.QueryInterface(Ci.nsIPrefBranch2);
    this._branch.removeObserver("", this);
    this._branch = null;
  },
  observe: function (aSubject, aTopic, aData) {
    if(aTopic == "nsPref:changed") {
      this._branch.QueryInterface(Ci.nsIPrefBranch);
      var autoUpload = this._branch.getBoolPref("");

      if(autoUpload)
        this._registerCloseObserver();
      else
        this._unregisterCloseObserver();
    } else if(aTopic == "domwindowclosed") {
      this._doLastUpload(aSubject);
    }
  },

  //_didLastUpload: false,
  _doLastUpload: function(aWindow) {
    //if(this._didLastUpload) // the closed window was the upload window
      //return;

    var windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"]
                           .getService(Ci.nsIWindowMediator);
    var win = windowMediator.getMostRecentWindow("");
    if(win) // there's at least one remaining window open
      return;

    //this._didLastUpload = true;
    this._unregisterCloseObserver();

    if(!this.dataChanged) // we don't have to do anything
      return;

    const URL = "chrome://adblockfilters/content/ftp-upload.xul";
    var up = new UploadJob();
    up.origWindow = aWindow;
    up.hash = this.hash;
    up.data = this.data;
    aWindow.openDialog(URL, "", "chrome", up, true);
  },

  isToolkitApp: function() {
    var appInfo = Cc["@mozilla.org/xre/app-info;1"]
                    .createInstance(Ci.nsIXULAppInfo);
    switch(appInfo.ID) {
      case "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}": // SeaMonkey
        var versionChecker = Cc["@mozilla.org/xpcom/version-comparator;1"]
                               .getService(Ci.nsIVersionComparator);
        if(versionChecker.compare("1.5a", appInfo.version) >= 0) {
           return ("nsIExtensionManager" in Ci) &&
                  ("@mozilla.org/extensions/manager;1" in Cc);
        }
        return false;
        break;
      case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": // Firefox
      case "{a463f10c-3994-11da-9945-000d60ca027b}": // Flock
      case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}": // Netscape
      case "{3550f703-e582-4d05-9a08-453d09bdfdc6}": // Thunderbird
      case "{718e30fb-e89b-41dd-9da7-e25a45638b28}": // Sunbird
      case "{136c295a-4a5a-41cf-bf24-5cee526720d5}": // NVU
      case "{758DAD28-FDE9-4ab8-A301-3FFFAB3A697A}": // Songbird
      case "songbird@songbirdnest.com":              // Songbird as well
      default:
        return true;
    }
  },
  openURL: function openURL(URL) {
    var appInfo = Cc["@mozilla.org/xre/app-info;1"]
                    .createInstance(Ci.nsIXULAppInfo);
    switch(appInfo.ID) {
      case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": // Firefox
      case "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}": // SeaMonkey
      case "{a463f10c-3994-11da-9945-000d60ca027b}": // Flock
      case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}": // Netscape
        var win = null;

        win = Cc["@mozilla.org/appshell/window-mediator;1"]
                .getService(Ci.nsIWindowMediator)
                .getMostRecentWindow("navigator:browser");

        const TAB_PREF = "browser.link.open_external";
        var tabsEnabled = nsPreferences.getIntPref(TAB_PREF) == 3;

        if(tabsEnabled && win) {
          win.focus();
          var browser = win.getBrowser();
          if(browser.selectedBrowser.currentURI.spec != "about:config")
            browser.selectedTab = browser.addTab(URL);
          else
            browser.loadURI(makeURI(URL));
        } else {
          var windowFeatures = "chrome=0,resizable,modal=0,dialog=0,close,"
                              + "minimizable,titlebar,menubar,toolbar,location,"
                              + "directories,personalbar,status";
          window.open(URL, "_blank", windowFeatures);
        }
        break;
      case "{3550f703-e582-4d05-9a08-453d09bdfdc6}": // Thunderbird
      case "{718e30fb-e89b-41dd-9da7-e25a45638b28}": // Sunbird
      case "{136c295a-4a5a-41cf-bf24-5cee526720d5}": // NVU
      case "{758DAD28-FDE9-4ab8-A301-3FFFAB3A697A}": // Songbird
      case "songbird@songbirdnest.com":              // Songbird as well
      default: // open in OS default
        Cc["@mozilla.org/uriloader/external-protocol-service;1"]
          .getService(Ci.nsIExternalProtocolService).loadUrl(makeURI(URL));
        break;
    }
  },

  get data() {
    var abp = Cc["@mozilla.org/adblockplus;1"].createInstance();
    while(abp && !("getString" in abp))
      abp = abp.wrappedJSObject;    // Unwrap component

    var prefs = abp.prefs;

    var text, data = [], abpRequired = false;
    for(var i = 0; i < prefs.userPatterns.length; i++) {

      text = prefs.userPatterns[i].text;

      if(prefs.userPatterns[i].disabled)
        text = "!" + text;

      data.push(text);
      switch(prefs.userPatterns[i].type) {
        case "whitelist":
        case "elemhide":
          abpRequired = true;
          break;
        case "filterlist":
          if(text[0] == "|")
            abpRequired = true;
          else if(text[text.length - 1] == "|")
            abpRequired = true;
          else if(text.indexOf("$") > -1)
            abpRequired = true;
          break;
      }
    }

    return "[Adblock" + (abpRequired ? " Plus" : "") + "]\n" + data.join("\n");
  },
  get hash() {
    var converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                      .createInstance(Ci.nsIScriptableUnicodeConverter);

    converter.charset = "UTF-8";
    var data = converter.convertToByteArray(this.data, {});

    var ch = Cc["@mozilla.org/security/hash;1"]
               .createInstance(Ci.nsICryptoHash);
    ch.init(ch.SHA1);
    ch.update(data, data.length);
    var hash = ch.finish(false);

    // convert the binary hash data to a hex string
    var hexString = "";
    for(var i in hash)
      hexString += ("0" + hash.charCodeAt(i).toString(16)).slice(-2);

    return hexString;
  },
  get dataChanged() {
    const PREF = "extensions.adblockfilters.lastUploadHash";
    var hash = this.hash;
    return (nsPreferences.copyUnicharPref(PREF) != hash);
  },

  get abpVersion() {
    if("nsIAdblockPlus" in Ci)
      return Cc["@mozilla.org/adblockplus;1"]
               .createInstance(Ci.nsIAdblockPlus).getInstalledVersion();
    else
      return "0";
  },


  // nsISupports
  QueryInterface: function(aIID) {
    if(aIID.equals(Ci.nsIWeakReference))
      throw Cr.NS_ERROR_NO_INTERFACE;
    return this;
  },

  // nsIClassInfo
  getHelperForLanguage: function() {
    return null;
  },
  getInterfaces: function getInterfaces(aCount) {
    var array = [Ci.nsIClassInfo];
    aCount.value = array.length;
    return array;
  },
  classDescription: "some helpers for the Adblock Plus Filter Uploader extension... nothing you could need",
  classIDNoAlloc: Ci.nsISupports,
  flags: Ci.nsIClassInfo.SINGLETON,
  implementationLanguage: Ci.nsIProgrammingLanguage.JAVASCRIPT
};


/*
 * Provides data and target URI to upload to.
 */
function UploadJob() {
  this.wrappedJSObject = this;
}
UploadJob.prototype = {
  classID: Components.ID("{3e0cb45d-4e51-fdf2-839b-5158df0c4d15}"),
  contractID: "@adblockfilters.mozdev.org/uploadjob;1",
  classDescription: "Adblock Plus Filter Uploader upload job",

  wrappedJSObject: null,

  get targetURI() {
    const PREF_IX = "extensions.adblockfilters.ftp.";
    var URI = Cc["@mozilla.org/network/standard-url;1"]
                .createInstance(Ci.nsIURI);

    var host = nsPreferences.copyUnicharPref(PREF_IX + "host");
    var path = nsPreferences.copyUnicharPref(PREF_IX + "path");
    var username = nsPreferences.copyUnicharPref(PREF_IX + "user");
    var protocol = nsPreferences.copyUnicharPref(PREF_IX + "protocol");//xxx ascii
    var password = "";

    if(!host || !path)
      throw new Error();; //xxx

    if(path[0] != "/") // xxx?
      path = "/" + path;

    try {
      // This doesn't harm if passwords are not encrypted
      var tokendb = Cc["@mozilla.org/security/pk11tokendb;1"]
                      .createInstance(Ci.nsIPK11TokenDB);
      var token = tokendb.getInternalKeyToken();

      if(!token.checkPassword("")) {
        // So there's a master password. But since checkPassword didn't
        // succeed, we're logged out (per nsIPK11Token.idl).
        // Relogin and ask for the master password.
        token.login(true);  // 'true' means always prompt for token password.
                            // User will be prompted until clicking 'Cancel'
                            // or entering the correct password.

        // An exception will be thrown if the user cancels the login prompt
        // dialog. User is also logged out of Software Security Device.
      }

      var passwordManager = Cc["@mozilla.org/passwordmanager;1"]
                              .createInstance(Ci.nsIPasswordManager);
      var passwords = passwordManager.enumerator;
      while(passwords.hasMoreElements()) {
        var nsIPassword = passwords.getNext();
        nsIPassword.QueryInterface(Ci.nsIPassword);
        if(nsIPassword.host == host && nsIPassword.user == username) {
          password = nsIPassword.password;
          break;
        }
      }
    } catch(e) {
      var prompts = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                      .getService(Ci.nsIPromptService);

      var title = "Adblock Plus Filter Uploader";

      username = { value: username };
      password = { value: password };
      if(!prompts.promptUsernameAndPassword(window, title,
                         "Server Authentication", username, password, null, {}))
        throw new Error("");
      username = username.value;
      password = password.value;
    }

    URI.spec = protocol + "://" + username + ":" + password + "@" + host + path;
    return URI;
  },

  get streamToUpload() {
    var storageStream = Components.classes["@mozilla.org/storagestream;1"]
                                  .createInstance(Components.interfaces
                                                         .nsIStorageStream);
    storageStream.init(8192, -1, null);
    var outputStream = storageStream.getOutputStream(0);

    try {
      var os = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                 .createInstance(Components.interfaces.nsIConverterOutputStream);
      os.init(outputStream, null, 0, "?".charCodeAt(0));
      os.writeString(this.data);
      os.flush();
      os.close();
    } catch(e) {
      Components.reportError(e);
    }

    outputStream.close();
    return storageStream.newInputStream(0);
  },

  contentType: "text/plain",

  data: null,
  hash: null,
  origWindow: null,
  onUploaded: function onUploaded() {
    const PREF = "extensions.adblockfilters.lastUploadHash";

    // for some reason the Components object isn't defined inside the function
    var prefs = Cc["@mozilla.org/preferences-service;1"]
                  .getService(Ci.nsIPrefBranch);
    var origWindow = this.origWindow;
    var str = Cc["@mozilla.org/supports-string;1"]
                .createInstance(Ci.nsISupportsString);
    str.data = this.hash;
    prefs.setComplexValue(PREF, Ci.nsISupportsString, str);

    if(origWindow && !origWindow.closed)
      origWindow.document.documentElement.getButton("extra1").disabled = true;
  },

  // nsISupports
  QueryInterface: function(aIID) {
    if(!aIID.equals(Ci.nsISupports))
      throw Cr.NS_ERROR_NO_INTERFACE;
    return this;
  }
};


// constructors for objects we want to XPCOMify
var objects = [AdblockFilterService, UploadJob];


/*
 * Common registration code.
 *
 */


function FactoryHolder(aObj) {
  this.CID        = aObj.prototype.classID;
  this.contractID = aObj.prototype.contractID;
  this.className  = aObj.prototype.classDescription;
  this.factory = {
    createInstance: function(aOuter, aIID) {
      if(aOuter)
        throw Cr.NS_ERROR_NO_AGGREGATION;

      return (new this.constructor).QueryInterface(aIID);
    }
  };
  this.factory.constructor = aObj;
}

var gModule = {
  registerSelf: function (aComponentManager, aFileSpec, aLocation, aType)
  {
    aComponentManager.QueryInterface(Ci.nsIComponentRegistrar);
    for (var key in this._objects) {
      var obj = this._objects[key];
      aComponentManager.registerFactoryLocation(obj.CID, obj.className,
        obj.contractID, aFileSpec, aLocation, aType);
    }
  },

  unregisterSelf: function(aCompMgr, aFileSpec, aLocation) {
    aComponentManager.QueryInterface(Ci.nsIComponentRegistrar);
    for (var key in this._objects) {
      var obj = this._objects[key];
      aComponentManager.unregisterFactoryLocation(obj.CID, aFileSpec);
    }
  },

  getClassObject: function(aComponentManager, aCID, aIID) {
    if (!aIID.equals(Ci.nsIFactory)) throw CR.NS_ERROR_NOT_IMPLEMENTED;

    for (var key in this._objects) {
      if (aCID.equals(this._objects[key].CID))
        return this._objects[key].factory;
    }

    throw CR.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aComponentManager) {
    return true;
  },

  _objects: {} //FactoryHolder
};

function NSGetModule(compMgr, fileSpec)
{
  for(var i in objects)
    gModule._objects[i] = new FactoryHolder(objects[i]);
  return gModule;
}